/************************************************************************/
/*  Routines for writing records to file.				*/
/*									*/
/*	Douglas Neuhauser						*/
/*	Seismological Laboratory					*/
/*	University of California, Berkeley				*/
/*	doug@seismo.berkeley.edu					*/
/*									*/
/************************************************************************/

/*
 * Copyright (c) 1996-2003 The Regents of the University of California.
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes,
 * without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice, this paragraph and the
 * following three paragraphs appear in all copies.
 * 
 * Permission to incorporate this software into commercial products may
 * be obtained from the Office of Technology Licensing, 2150 Shattuck
 * Avenue, Suite 510, Berkeley, CA  94704.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
 * ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 * CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#ifndef lint
static char sccsid[] = "$Id: write_file2.c,v 1.2 2009/03/12 12:41:17 mario Exp $ ";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <utime.h>

#include "qlib2.h"
#include "params.h"
#include "defs.h"
#include "st_info.h"
#include "procs.h"
#include "externals.h"


/************************************************************************/
/*  do_write_file2:							*/
/*	Write output file.  Currently supports only DRM and SDR files.	*/
/************************************************************************/
void do_write_file2
   (ST_INFO	*st_p,		/* ptr to input stream struct.		*/
    BLKINFO	**blks,		/* ptr to array of BLKINFO ptrs.        */
    char	*filename)	/* output filename.			*/
{
    ST_INFO	*out;
    BLKINFO	*bp;

    struct stat sbuf;
    time_t save_time;
    struct utimbuf utb;
    
    if (filename == NULL) {
        if(!outoforder) {
           if (debug(DEBUG_CMD)) {
	      fprintf (info, "Not overwriting input file\n");
	   }
	   return;
        }
        if (debug(DEBUG_CMD)) {
	   fprintf (info, "Overwriting input file: %s\n", st_p->filename);
	}
        filename = st_p->filename;
	if ((out = open_stream2(&st_head_out, st_p, filename, "default")) == NULL) {
	   fprintf (info, "Unable to open output file %s\n", filename);
	   return;
	}
    } else {
        if ((out = open_stream(&st_head_out, filename, "w")) == NULL) {
    	   fprintf (info, "Unable to open output file %s\n", filename);
	   return;
        }
    } 

    if (debug(DEBUG_CMD)) {
        fprintf (info, "Writing file: %s\n", filename);
    }

    /*	For the moment, don't even bother with any BLKINFO, since all	*/
    /*	we need to do is write to our iob.				*/
    /*	We need a cur_hdr for summary info, however.			*/
    bp = st_p->cur;
    out->cur_hdr = new_data_hdr();

    if( out->iob->fd == fileno(stdin) || out->iob->fd == fileno(stdout) ) {
      preserve_time = 0;
    }
					   
    if(preserve_time) {
       if(fstat (st_p->iob->fd, &sbuf) != 0) {
	  fprintf(info, "Error: cannot fstat input file %s\n", st_p->filename);
	  exit(1);
       }
       save_time = sbuf.st_mtime;
    }

    switch (st_p->hdr_type) {
    case DRM_HDR_TYPE:
	write_file2_drm (st_p, blks, out); break;
    case QDA_HDR_TYPE:
	/*  Currently let this default to writing SDR output files.	*/
    case SDR_HDR_TYPE:
	write_file2_sdr (st_p, blks, out); break;
    default:
	fprintf (stderr, "unable to write file -- Unknown header type\n");
    }

    close_stream(out);
    st_p->cur = bp;

    /* restore file mod time */
    if(preserve_time) {
      if(stat (out->filename, &sbuf) != 0) {
	  fprintf(info, "Error: cannot fstat output file %s\n", out->filename);
	  exit(1);
      }
      utb.actime = sbuf.st_atime;
      utb.modtime = save_time;
      utime(out->filename, &utb);
    }

    return;
}

/************************************************************************/
/*  write_file2_sdr:							*/
/*	Write file in SDR format.					*/
/************************************************************************/
void write_file2_sdr
   (ST_INFO	*st_p,		/* ptr to input stream structure.	*/
    BLKINFO	**blks,		/* ptr to array of BLKINFO ptrs.        */
    ST_INFO	*out)		/* ptr to output stream structure.	*/
{
    int		o_offset, datalen, i;

    /*	For the moment, write out SEED data records with the same	*/
    /*	blocksize as the input file.					*/
    /*  Write out new volume header if one existed before.		*/
    st_p->cur = blks[0];
    write_vol(st_p, out);

    for(i=0; i < st_p->num_blocks; i++) {
        st_p->cur = blks[i];
	memset ((void *)out->iob->buf, 0, st_p->blksize);
	init_sdr_hdr ((SDR_HDR *)out->iob->buf, st_p->cur->hdr, NULL);
	o_offset = st_p->cur->hdr->first_data;
	datalen = st_p->cur->datalen;
	out->blksize = st_p->blksize;
	if (datalen > 0) {
	    memcpy (out->iob->buf+o_offset, st_p->cur->data, datalen);
	}
	update_sdr_hdr ((SDR_HDR *)out->iob->buf, st_p->cur->hdr);
	write_stream (out);
    }
}


/************************************************************************/
/*  write_file2_drm:							*/
/*	Write file in DRM format.					*/
/************************************************************************/
void write_file2_drm
   (ST_INFO	*st_p,		/* ptr to input stream structure.	*/
    BLKINFO	**blks,		/* ptr to array of BLKINFO ptrs.        */
    ST_INFO	*out)		/* ptr to output stream structure.	*/
{
    char	*chan = NULL;
    char	*stat = NULL;
    char	*loc = NULL;
    int		n_blocks = 0;
    STORE_DATA	*sdh;
    int		o_offset, datalen, i;

    /*  Output the file config hearder.  Ensure all info is correct.	*/
    /*	Certain info such as station, channel, and location MUST be	*/
    /*	consistent throughout the file, since they are represented only	*/
    /*	once in the file header.					*/
    if ((st_p->cur = blks[0]) && st_p->cur->hdr) {
	stat = st_p->cur->hdr->station_id;
	chan = st_p->cur->hdr->channel_id;
	loc = st_p->cur->hdr->location_id;
    }
    if (st_p->cur == NULL) {
	fprintf (info, "no datablocks for file, unable to write\n");
	return;
    }

    for(i=0; i < st_p->num_blocks; i++) {
        st_p->cur = blks[i];
	if (strcmp(stat, st_p->cur->hdr->station_id)) {
	    fprintf (info, "station mismatch at block %d, seq %d\n", 
		     st_p->cur->blk_no, st_p->cur->hdr->seq_no);
	    return;
	}
	if (strcmp(chan, st_p->cur->hdr->channel_id)) {
	    fprintf (info, "channel mismatch at block %d, seq %d\n", 
		     st_p->cur->blk_no, st_p->cur->hdr->seq_no);
	    return;
	}
	if (strcmp(loc, st_p->cur->hdr->location_id)) {
	    fprintf (info, "location mismatch at block %d, seq %d\n", 
		     st_p->cur->blk_no, st_p->cur->hdr->seq_no);
	    return;
	}
	++n_blocks;
    }
    /*	Update file header information and write file header.		*/
    /*	The file header uses an integral number of packets.		*/
    strncpy (st_p->sfh->stream_config.seedchannel,chan,3);
    strncpy (st_p->sfh->stream_config.station_name,stat,4);
    st_p->sfh->store_count = n_blocks;
    st_p->sfh->store_begin = encode_time_drm(blks[0]->hdr->begtime, 
					     SEED_BIG_ENDIAN);
    st_p->sfh->store_end = encode_time_drm(blks[st_p->num_blocks]->hdr->endtime,
					   SEED_BIG_ENDIAN);
    out->blksize = st_p->sfh->store_pack_size * st_p->sfh->store_offset;
    memset(out->iob->buf, 0, out->blksize);
    memcpy (out->iob->buf, st_p->sfh, sizeof(STORE_FILE_HEAD));
    write_stream (out);

    /*	Convert each block back to a drm block, and write block.	*/
    out->blksize = st_p->sfh->store_pack_size;
    for(i=0; i < st_p->num_blocks; i++) {
        st_p->cur = blks[i];
	memset ((void *)out->iob->buf, 0, st_p->blksize);
	sdh = encode_hdr_drm(st_p->cur->hdr);
	o_offset = (char *)&(sdh->da_d[0][0]) - (char *)sdh;
	/*  Copy only the header portion of the STORE_DATA.		*/
	memcpy (out->iob->buf, sdh, o_offset);
	datalen = st_p->cur->datalen;
	memcpy (out->iob->buf+o_offset, st_p->cur->data, datalen);
	out->blksize = o_offset + datalen;
	write_stream (out);
    }
}
